/** =======================================================================
 *  Vincent Hardy
 *  License terms: see svg-wow.org
 *  CC0 http://creativecommons.org/publicdomain/zero/1.0/
 *  ======================================================================= */
(function () {
    /**
     * YUI global object instance.
     * See the load call at the end of this file.
     */
    var Y;
    
    /**
     * Utility to get the viewbox information.
     */
    function getViewBox () {
        var svg = Y.one('svg'), vb = svg.get('viewBox');
        return {
            x: vb.baseVal.x,
            y: vb.baseVal.y,
            width: vb.baseVal.width,
            height: vb.baseVal.height
        };
    }

    /**
     * The size of each polaroid picture frame.
     */
    var polaroidSize = {width: 80, height: 80};

    /**
     * The polaroidSpacing gives the amount of space along each axis between
     * the polaroid's anchor point.
     */
    var polaroidSpacing = {width: 60, height: 60};

    /**
     * The amount of overlap between polaroid pictures in the [0, 1] range.
     * The allows the polaroids to move about their anchor point in a random
     * way, bounded by the min and max overlap values.
     */
    var overlap = {min: 0, max: 0.2};

    /**
     * The amount of rotation for the polaroid pictures about their anchor point.
     */
    var rotation = {min: 5, max: 90};

    /**
     * The amount of randomness given to the alignment between polaroids.
     * 0 is no randomness at all. 1 is the maximum randomness allowed.
     */
    var alignmentRoughness = 0.01; // +/-1% of random alignment

    /**
     * The margin around the image
     */
    var margin = 100;

    /**
     * The polaroid border width
     */
    var borderWidth = 4;

    function onScriptsReady (YUI) {
        Y = YUI;
        createPolaroidStack('magasin.png',
                            {x: 200, y: 600},
                            positionAndAnimateStack);
        createPolaroidStack('bretagne2.png',
                            {x: 400, y: 600},
                            positionAndAnimateStack);
        createPolaroidStack('romain.png',
                            {x: 600, y: 600},
                            positionAndAnimateStack);
    }

    /**
     * Creates a stack of images in their final position.
     *
     * @param imageURL the image to display.
     * @param position the desired stack position in viewBox space.
     * @param callback the function to invoke when the stack has been created.
     * @return an SVG group with the polaroid elements.
     */
    function createPolaroidStack (imageURL, position, callback) {
        var img = new Image();

        img.onload = function () {
            callback(onPolaroidStackSourceReady(img), position);
        };

        img.onerror = function () {
            alert('Could not load image ' + imageURL);
        };

        // Setting the url triggers the image load.
        img.src = imageURL;
    }

    /**
     * Positions and animates the polaroid stack.
     *
     * @param stack the stack of polaroids
     * @param position the desired stack position, in viewBox space.
     */
    function positionAndAnimateStack (stack, position) {
        var c = stack.get('lastChild'),
            anims = [], anim = null, prevAnim = null,
            stackPosition, stackTxf, stackTop = c;

        // The stack position is in the stack space. Get the stack's
        // transform and combine its translation component with the desired
        // position.
        stackTxf = stack.getMatrix();
        stackPosition = {
            x: position.x - stackTxf.e(),
            y: position.y - stackTxf.f()
        };

        function getReverseHandler (a) {
            return function () {
                a.set('reverse', !a.get('reverse'));
                if (a.get('reverse') === true) {
                    a.set('duration', 0.15 + 0.15 * Math.random());
                } else {
                    a.set('duration', 0.25 + 0.25 * Math.random());
                }
            };
        }

        while (c !== null) {
            prevAnim = anim;
            anim = new sw.animation.Animate({
                node: c,
                from: {
                    transform: {
                        r: c.r + 90,
                        tx: stackPosition.x + 10 - 20 * Math.random(),
                        ty: stackPosition.y + 10 - 10 * Math.random()
                    }
                },
                to: {
                    transform: {r: c.r, tx: c.cx, ty: c.cy}
                },
                transformTemplate: "translate(#tx, #ty) rotate(#r) " +
                                   "translate(" + (-c.cx) + "," + (-c.cy) +
                                   ")",
                duration: 0.25 + 0.25 * Math.random(),
                easing: Y.Easing.easeOutStrong
            });
            anims.push(anim);
            anim.applyStartFrame();
            if (prevAnim !== null) {
                prevAnim.onBegin(anim, 0.05 * Math.random());
            }

            anim.onEnd(getReverseHandler(anim));
            c = c.get('previousSibling');
        }

        Y.one('#polaroidStacks').appendChild(stack);
        anims[0].beginOn(stack, 'click');

        stackTop.trace = true;
    }

    /**
     * Called when an image has been successfully loaded.
     *
     * @param img the image to use as the source for the stack.
     */
    function onPolaroidStackSourceReady(img) {
        var viewBox = getViewBox(),
            imageWidth, imageHeight,
            nCols, nRows,
            polaroidStack = Y.one(document.createElementNS(Y.svgNS, 'g')),
            sx, sy, s,
            cx, cy, r,
            ox, oy,
            dr = rotation.max - rotation.min,
            ds = overlap.max - overlap.min,
            polaroid,
            allPictures = [],
            prev = null,
            canvas, ctx,
            shadowBlur = 6,
            shadowOffsetX = 0,
            shadowOffsetY = 0,
            shadowColor = 'rgba(0,0,0,0.85)',
            shadowPadding = {
                width: 5,
                height: 5
            },
            snapshotSize = {
                width: polaroidSize.width + 2 * shadowPadding.width,
                height: polaroidSize.height + 2 * shadowPadding.height
            },
            polaroidTemplate = Y.one(Y.DOM.loadContent(null, {
                tag: 'image',
                'class': 'polaroid',
                width: snapshotSize.width,
                height: snapshotSize.height
            }));

        // =====================================================================
        // We want to display the image in the viewBox with some margin.
        // To achieve that, we compute the necessary scale factor for the
        // image.
        // =====================================================================

        // The padding allows rotated polaroids to still show the image content.
        // Without padding, a rotated polaroid on the edges would show part of
        // the image on the opposite side. For example, a polaroid on the
        // top would show content from the bottom.
        var picturePadding = {
            width: polaroidSize.width * (1 - 1 / Math.sqrt(2)) +
                   polaroidSize.width * overlap.max,
            height: polaroidSize.height * (1 - 1 / Math.sqrt(2)) +
                   polaroidSize.height * overlap.max
        };

        sx = (viewBox.width - margin * 2) / img.width;
        sy = (viewBox.height - margin * 2) / img.height;

        s = sx;
        if (sx > sy) {
            s = sy;
        }

        imageWidth = Math.round(img.width * s);
        imageHeight = Math.round(img.height * s);

        // Make padding conform to the image's aspect ratio
        var ar = img.width / img.height;
        picturePadding.width *= ar;

        var pictureWidth = imageWidth + 2 * picturePadding.width;
        var pictureHeight = imageHeight + 2 * picturePadding.height;


        // =====================================================================
        // Create a canvas where we'll draw the tiles as we need them.
        // =====================================================================
        canvas = document.createElement('canvas');
        canvas.width = snapshotSize.width;
        canvas.height = snapshotSize.height;
        ctx = canvas.getContext('2d');
        ctx.save();

        // imgCanvas is used to hold the scaled image of the source image.
        // We scale the image a bit compared to the target size to account for
        // the polaroids rotation and make sure they do not bleed over the
        // picture's edges.
        var imgCanvas = document.createElement('canvas');
        imgCanvas.width = pictureWidth;
        imgCanvas.height = pictureHeight;
        imgCanvas.getContext('2d').drawImage(img, 0, 0,
                     img.width, img.height, 0, 0, pictureWidth, pictureHeight);


        // snapshotCanvas is used to take a rotated snapshot to include in a
        // polaroid rendering.
        var snapshotCanvas = document.createElement('canvas');
        snapshotCanvas.width = polaroidSize.width;
        snapshotCanvas.height = polaroidSize.height;
        var snapshotCtx = snapshotCanvas.getContext('2d');

        /**
         * Private method.
         * Takes a picture of the polaroid centered on cx/cy which will be
         * rotated by r degrees.
         *
         * The resulting image is stored in snapshotCanvas.
         */
        function snapshot (cx, cy, r) {
            var p = alignmentRoughness;
            var ar = (1 - p + 2 * p * Math.random()) * r,
                acx = (1 - p + 2 * p * Math.random()) * cx,
                acy = (1 - p + 2 * p * Math.random()) * cy;

            snapshotCtx.save();
            snapshotCtx.clearRect(0, 0, polaroidSize.width, polaroidSize.height);
            snapshotCtx.translate(polaroidSize.width / 2, polaroidSize.height /2);
            snapshotCtx.rotate(-ar * Math.PI / 180);
            snapshotCtx.translate(-acx, -acy);
            snapshotCtx.drawImage(imgCanvas, 0, 0);
            snapshotCtx.restore();
        }
        
        /**
         * Private method.
         * Takes a picture of the polaroid centered on cx/cy and add decoration
         * to it (drop shadow and border).
         *
         * @return a string dataURL with the polaroid's image.
         */
        function polaroidSnapshot (cx, cy, r) {
            // Create the snapshot we need
            snapshot(cx, cy, r);

            ctx.clearRect(0, 0, snapshotSize.width, snapshotSize.height);
            ctx.shadowBlur = shadowBlur;
            ctx.shadowOffsetX = shadowOffsetX;
            ctx.shadowOffsetY = shadowOffsetY;
            ctx.shadowColor = shadowColor;
            ctx.drawImage(snapshotCanvas, shadowPadding.width, shadowPadding.height);
            ctx.restore();
            ctx.strokeStyle = '#fff';
            ctx.lineWidth = borderWidth;
            ctx.strokeRect(shadowPadding.width + borderWidth / 2,
                           shadowPadding.height + borderWidth / 2,
                           polaroidSize.width - borderWidth,
                           polaroidSize.height - borderWidth);
            ctx.save();
            
            return canvas.toDataURL();
        }

        // =====================================================================
        // Create the polaroids.
        // =====================================================================
        nCols = Math.floor(imageWidth / polaroidSpacing.width);
        nRows = Math.floor(imageHeight / polaroidSpacing.height);

        var dx = 0;
        if (nCols > 0) {
            dx = (imageWidth - nCols * polaroidSpacing.width) / 2;
        }

        var dy = 0;
        if (nRows > 0) {
            dy = (imageHeight - nRows * polaroidSpacing.height) / 2;
        }

        polaroidStack.set('transform', 'translate(' +
            (viewBox.width - imageWidth) / 2 + ',' +
            (viewBox.height - imageHeight) / 2 + ')');

        for (var ri = 0; ri < nRows; ri++) {
            for (var ci = 0; ci < nCols; ci++) {
                ox = overlap.min + ds * Math.random();
                if (Math.random() > 0.5) {
                    ox *= -1;
                }

                oy = overlap.min + ds * Math.random();
                if (Math.random() > 0.5) {
                    oy *= -1;
                }

                // cx/cy is the center of the polaroid tile in the the polaroid
                // stack group space. The so-called picturePadding is offset
                // compared to the polaroid stack coordinate space by the
                // pattern padding.
                //
                // This is why we offset the polaroidSnapshot call below by
                // the picturePadding amount along both axis.
                cx = polaroidSpacing.width * (ci + 0.5 + ox) + dx;
                cy = polaroidSpacing.height * (ri + 0.5 + oy) + dy;
                r = rotation.min + Math.random() * dr;
                if (Math.random() > 0.5) {
                    r = -r;
                }

                polaroid = polaroidTemplate.cloneNode();

                polaroid.setAttrs({
                    id: "polaroid_" + ri + "_" + ci,
                    x: cx - snapshotSize.width / 2,
                    y: cy - snapshotSize.height / 2
                });

                polaroid._node.setAttributeNS(Y.xlinkNS, 'href',
                            polaroidSnapshot(cx + picturePadding.width,
                                             cy + picturePadding.height, r));

                polaroid.set('transform',
                                  'rotate(' + r + ',' + cx + ',' + cy + ')');
                polaroid.r = r;
                polaroid.cx = cx;
                polaroid.cy = cy;

                if (allPictures.length > 0) {
                    prev = allPictures[
                        Math.round((allPictures.length - 1) * Math.random())];
                }

                polaroidStack.insertBefore(polaroid, prev);
                allPictures.push(polaroid);
            }
        }

        return polaroidStack;
    }

    /**
     * Main entry point. Refer to the YUI documentation to understand how the
     * YUI framework loads scripts and handles dependencies.
     *
     * The following declares the modules required by the code (such as
     * 'anim-svg') and their dependencies.
     *
     * The onScriptsReady function is called when the DOM tree is ready.
     */
    YUI({
        groups: {
            svg: {
                combine: false,
                base: "../res/3.0/scripts/yui/",
                modules:  {
                    "node-svg": {
                        path: 'svg/node-svg-debug.js',
                        requires: ['node']
                    },
                    "dom-style-svg": {
                        path: 'svg/dom-style-svg-debug.js',
                        requires: ['node', 'dom-svg']
                    },
                    "dom-svg": {
                        path: 'svg/dom-svg-debug.js',
                        requires: ['dom']
                    },
                    "anim-timing": {
                        path: 'anim-timing/anim-timing-debug.js',
                        requires: ['anim', 'node-svg']
                    },
                    "anim-svg": {
                        path: 'svg/anim-svg-debug.js',
                        requires: ['anim', 'dom-style-svg']
                    },
                    "text-effects": {
                        path: '../tools/text-effects.js',
                        requires: ['anim-svg', 'anim-timing']
                    }
                }
            }
        }
    }).use('anim-svg', 'anim-timing', 'text-effects', 'dom-utils', function(YUI) {
        YUI.on('contentready', function () {
            onScriptsReady(YUI);
        }, 'body');
    });
    
})();